# License: BSD 3 clause
"""
A meta-learner class that wraps scikit-learn's `VotingClassifier` and `VotingRegressor`.

:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""
import copy
import logging
from importlib import import_module
from itertools import zip_longest
from multiprocessing import cpu_count
from typing import Any, Dict, List, Optional, Tuple, Union

import joblib
import numpy as np
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.utils import shuffle as sk_shuffle
from sklearn.utils.multiclass import type_of_target

from skll.data import FeatureSet
from skll.data.dict_vectorizer import DictVectorizer
from skll.learner import Learner
from skll.types import (
    EvaluateTaskResults,
    FoldMapping,
    LabelType,
    LearningCurveSizes,
    PathOrStr,
    VotingCrossValidateTaskResults,
)
from skll.utils.constants import MAX_CONCURRENT_PROCESSES

from .utils import (
    _load_learner_from_disk,
    _save_learner_to_disk,
    add_unseen_labels,
    compute_evaluation_metrics,
    get_acceptable_classification_metrics,
    get_acceptable_regression_metrics,
    get_predictions,
    setup_cv_fold_iterator,
    setup_cv_split_iterator,
    train_and_score,
    write_predictions,
)


class VotingLearner(object):
    """
    Wrap ``VotingClassifier`` and ``VotingRegressor`` from scikit-learn.

    Note that this class does not inherit from the ``Learner`` class but rather
    uses different ``Learner`` instances underlyingly.

    Parameters
    ----------
    learner_names : List[str]
        List of the learner names that will participate in the voting process.
    voting : Optional[str], default="hard"
        One of "hard" or "soft". If "hard", the predicted class labels
        are used for majority rule voting. If "soft", the predicted class
        label is based on the argmax of the sums of the predicted
        probabilities from each of the underlying learnrs. This parameter
        is only relevant for classification.
    custom_learner_path : Optional[:class:`skll.types.PathOrStr`], default=None
        Path to a Python file containing the definitions of any custom
        learners. Any and all custom learners in ``estimator_names`` must
        be defined in this file. If the custom learner does not inherit
        from an already existing scikit-learn estimator, it must explicitly
        define an `_estimator_type` attribute indicating whether it's a
        "classifier" or a "regressor".
    feature_scaling : str, default="none"
        How to scale the features, if at all for each estimator. Options are
        -  "with_std": scale features using the standard deviation
        -  "with_mean": center features using the mean
        -  "both": do both scaling as well as centering
        -  "none": do neither scaling nor centering
    pos_label : Optional[:class:`skll.types.LabelType`], default=None
        A string denoting the label of the class to be
        treated as the positive class in a binary classification
        setting, for each estimator. If ``None``, the class represented
        by the label that appears second when sorted is chosen as the
        positive class. For example, if the two labels in data are "A"
        and "B" and ``pos_label`` is not specified, "B" will
        be chosen as the positive class.
    min_feature_count : int, default=1
        The minimum number of examples a feature must have a nonzero
        value in to be included, for each estimator.
    model_kwargs_list : Optional[List[Dict[str, Any]]], default=None
        A list of dictionaries of keyword arguments to pass to the
        initializer for each of the estimators. There's a one-to-one
        correspondence between the order of this list and the order
        of the ``learner_names`` list.
    sampler_list : Optional[List[str]], default=None
        The samplers to use for kernel approximation, if desired, for each
        estimator. Valid values are:
        -  "AdditiveChi2Sampler"
        -  "Nystroem"
        -  "RBFSampler"
        -  "SkewedChi2Sampler"
        There's a one-to-one correspondence between the order of this list
        and the order of the ``learner_names`` list.
    sampler_kwargs_list : Optional[List[Dict[str, Any]]], default=None
        A list of dictionaries of keyword arguments to pass to the
        initializer for the specified sampler, one per estimator.
        There's a one-to-one correspondence between the order of this
        list and the order of the ``learner_names`` list.
    logger : Optional[logging.Logger], default=None
        A logging object. If ``None`` is passed, get logger from ``__name__``.
    """

    def __init__(
        self,
        learner_names: List[str],
        voting: Optional[str] = "hard",
        custom_learner_path: Optional[PathOrStr] = None,
        feature_scaling: str = "none",
        pos_label: Optional[LabelType] = None,
        min_feature_count: int = 1,
        model_kwargs_list: Optional[List[Dict[str, Any]]] = None,
        sampler_list: Optional[List[str]] = None,
        sampler_kwargs_list: Optional[List[Dict[str, Any]]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize a ``VotingLearner`` object with the specified settings."""
        # initialize various attributes
        self._model = None
        self.voting = voting
        self.label_dict: Dict[LabelType, int] = {}
        self.logger = logger if logger else logging.getLogger(__name__)
        self.model_kwargs_list = [] if model_kwargs_list is None else model_kwargs_list
        self.sampler_list = [] if sampler_list is None else sampler_list
        self.sampler_kwargs_list = [] if sampler_kwargs_list is None else sampler_kwargs_list

        # check that the arguments that are supposed to be lists are lists;
        # if they are `None`, set them to be empty lists
        for argument_name in ["model_kwargs_list", "sampler_list", "sampler_kwargs_list"]:
            argument_value = locals()[argument_name]
            if argument_value is None:
                setattr(self, argument_name, [])
            else:
                if not isinstance(argument_value, list):
                    raise ValueError(
                        f"{argument_name} should be a list, you " f"specified {argument_value}"
                    )
                else:
                    setattr(self, argument_name, argument_value)

        # check that the list arguments, if not empty, have the right length
        for attribute_name in ["model_kwargs_list", "sampler_list", "sampler_kwargs_list"]:
            attribute_value = getattr(self, attribute_name)
            try:
                assert len(attribute_value) == 0 or len(attribute_value) == len(learner_names)
            except AssertionError:
                raise ValueError(
                    f"'{attribute_name}' must have {len(learner_names)} "
                    "entries, same as the number of learners"
                ) from None

        # instantiate each of the given estimators
        self._learners = []
        learner_types = set()
        self._learner_names = learner_names
        for learner_name, model_kwargs, sampler, sampler_kwargs in zip_longest(
            self._learner_names, self.model_kwargs_list, self.sampler_list, self.sampler_kwargs_list
        ):
            learner = Learner(
                learner_name,
                custom_learner_path=custom_learner_path,
                feature_scaling=feature_scaling,
                min_feature_count=min_feature_count,
                model_kwargs=model_kwargs,
                pipeline=True,
                pos_label=pos_label,
                probability=self.voting == "soft",
                sampler=sampler,
                sampler_kwargs=sampler_kwargs,
                logger=logger,
            )
            learner_types.add(learner.model_type._estimator_type)
            self._learners.append(learner)

        # infer what type of metalearner we have - a classifier or
        # a regressor; it can only be one or the other
        try:
            assert len(learner_types) == 1 and (
                learner_types == {"classifier"} or learner_types == {"regressor"}
            )
        except AssertionError:
            raise ValueError("cannot mix classifiers and regressors for voting")
        else:
            self.learner_type = list(learner_types)[0]

        # unset the voting attribute for regressors for downstream simplicity
        if self.learner_type == "regressor":
            self.voting = None

    @property
    def learners(self) -> List[Learner]:
        """Return the underlying list of learners."""
        return self._learners

    @property
    def model(self):
        """Return underlying scikit-learn meta-estimator model."""
        return self._model

    @property
    def model_type(self):
        """Return meta-estimator model type (i.e., the class)."""
        return self._model_type

    def _setup_underlying_learners(self, examples: FeatureSet) -> None:
        """Complete pre-training set up for learners."""
        for learner in self.learners:
            learner._create_label_dict(examples)
            learner._train_setup(examples)

    def __getstate__(self) -> Dict[str, Any]:
        """
        Return attributes that should be pickled.

        We need this because we do not want to dump loggers.
        """
        attribute_dict = dict(self.__dict__)
        if "logger" in attribute_dict:
            del attribute_dict["logger"]
        return attribute_dict

    def save(self, learner_path: PathOrStr) -> None:
        """
        Save the ``VotingLearner`` instance to a file.

        Parameters
        ----------
        learner_path : :class:`skll.types.PathOrStr`
            The path to save the ``VotingLearner`` instance to.
        """
        _save_learner_to_disk(self, learner_path)

    @classmethod
    def from_file(
        cls, learner_path: PathOrStr, logger: Optional[logging.Logger] = None
    ) -> "VotingLearner":
        """
        Load a saved ``VotingLearner`` instance from a file.

        Parameters
        ----------
        learner_path : :class:`skll.types.PathOrStr`
            The path to a saved ``VotingLearner`` instance file.
        logger : Optional[logging.Logger], default=None
            A logging object. If ``None`` is passed, get logger from ``__name__``.

        Returns
        -------
        learner : skll.learner.voting.VotingLearner
            The ``VotingLearner`` instance loaded from the file.
        """
        # use the logger that's passed in or if nothing was passed in,
        # then create a new logger
        logger = logger if logger else logging.getLogger(__name__)

        # call the learner loding utility function
        obj = _load_learner_from_disk(cls, learner_path, logger)
        assert isinstance(obj, cls)
        return obj

    def train(
        self,
        examples: FeatureSet,
        param_grid_list: Optional[List[Dict[str, Any]]] = None,
        grid_search_folds: Union[int, FoldMapping] = 5,
        grid_search: bool = True,
        grid_objective: Optional[str] = None,
        grid_jobs: Optional[int] = None,
        shuffle: bool = False,
    ) -> None:
        """
        Train the voting meta-estimator.

        First, we train each of the underlying estimators (represented by
        a skll ``Learner``), possibly with grid search. Then, we instantiate
        a ``VotingClassifier`` or ``VotingRegressor`` as appropriate with the
        scikit-learn ``Pipeline`` stored in the ``pipeline`` attribute
        of each trained ``Learner`` instance as the estimator. Finally,
        we call ``fit()`` on the ``VotingClassifier`` or ``VotingRegressor``
        instance. We follow this process because it allows us to use grid
        search to find good hyperparameter values for our underlying learners
        before passing them to the meta-estimator AND because it allows us to
        use SKLL featuresets and do all of the same pre-processing when
        doing inference.

        The trained meta-estimator is saved in the ``_model`` attribute.
        Nothing is returned.

        Parameters
        ----------
        examples : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance to use for training.
        param_grid_list : Optional[List[Dict[str, Any]]], default=None
            The list of parameter grids to search through for grid
            search, one for each underlying learner. The order of
            the dictionaries should correspond to the order in
            which the underlying estimators were specified when the
            ``VotingLearner`` was instantiated. If ``None``, the default
            parameter grids will be used for the underlying estimators.
        grid_search_folds : Union[int, :class:`skll.types.FoldMapping`], default=5
            The number of folds to use when doing the grid search
            for each of the underlying learners, or a mapping from
            example IDs to folds.
        grid_search : bool, default=True
            Should we use grid search when training each underlying learner?
        grid_objective : Optional[str], default=None
            The name of the objective function to use when
            doing the grid search for each underlying learner.
            Must be specified if ``grid_search`` is ``True``.
        grid_jobs : Optional[int], default=None
            The number of jobs to run in parallel when doing the
            grid search for each underlying learner. If ``None`` or 0,
            the number of grid search folds will be used.
        shuffle : bool, default=False
            Shuffle examples (e.g., for grid search CV.)
        """
        if param_grid_list is None:
            self._param_grids = []
        else:
            if not isinstance(param_grid_list, list):
                raise ValueError(
                    f"`param_grid_list` should be a list of dictionaries, "
                    f"you specified: {param_grid_list}"
                )
            else:
                self._param_grids = param_grid_list

        # train each of the underlying estimators with grid search, if required;
        # basically, we are just running grid search to find good hyperparameter
        # values that we can then pass to scikit-learn below
        for learner, param_grid in zip_longest(self.learners, self._param_grids):
            _ = learner.train(
                examples,
                grid_search=grid_search,
                grid_objective=grid_objective,
                param_grid=param_grid,
                grid_search_folds=grid_search_folds,
                grid_jobs=grid_jobs,
                shuffle=shuffle,
            )

        # once we have our instantiated learners, we use their `pipeline`
        # attribute as the input estimators to the specific voting learner type
        estimators = list(zip(self._learner_names, [learner.pipeline for learner in self.learners]))
        if self.learner_type == "classifier":
            self._model_type = VotingClassifier
            model_kwargs = {"voting": self.voting}
        else:
            self._model_type = VotingRegressor
            model_kwargs = {}
        meta_learner = self.model_type(estimators, **model_kwargs)

        # get the training features in the right dictionary format
        if isinstance(examples.vectorizer, DictVectorizer):
            X_train = examples.vectorizer.inverse_transform(examples.features)

        # since label dictionaries are identical for all underlying
        # learners, save it into a easier to access attribute
        self.label_dict = self.learners[0].label_dict

        # get the training labels in the right format too
        # NOTE: technically, we could also use a `LabelEncoder` here but
        # that may not account for passing `pos_label` above when
        # instantiating the learners so we stick with the label dict
        y_train: np.ndarray
        if examples.labels is not None:
            if self.learner_type == "classifier":
                y_train = np.array([self.label_dict[label] for label in examples.labels])
            else:
                # for regressors, the labels are just the labels
                y_train = examples.labels

        # now we need to fit the actual meta learner which will also fit
        # clones of the underlying pipelines;
        # NOTE: this will *not* yield the same results as the case where we take
        # the predictions from the trained SKLL learners above and do the
        # voting ourselves. This is because SKLL learners do a lot of things
        # differently (e.g., shuffling before grid search) that can cause
        # differences in results and/or floating point precision.
        self._model = meta_learner.fit(X_train, y_train)

    def predict(
        self,
        examples: FeatureSet,
        prediction_prefix: Optional[str] = None,
        append: bool = False,
        class_labels: bool = True,
        individual_predictions: bool = False,
    ) -> Tuple[np.ndarray, Optional[Dict[str, np.ndarray]]]:
        """
        Generate predictions with meta-estimator.

        Compute the predictions from the meta-estimator and, optionally, the
        underlying estimators on given ``FeatureSet``. The predictions are
        also written to disk if ``prediction_prefix`` is not ``None``.

        For regressors, the returned and written-out predictions are identical.
        However, for classifiers:

        - if ``class_labels`` is ``True``, class labels are returned as well as
          written out.

        - if ``class_labels`` is ``False`` and the classifier is probabilistic
          (i.e., ``self.probability`` is ``True``), class probabilities are
          returned as well as written out.

        - if ``class_labels`` is ``False`` and the classifier is non-probabilistic
          (i.e., ``self..probability`` is ``False``), class indices are returned
          and class labels are written out. This option is generally only
          meant for SKLL-internal use.

        Parameters
        ----------
        examples : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance to predict labels for.
        prediction_prefix : Optional[str], default=None
            If saving the predictions, this is the prefix that will be used for
            the filename. It will be followed by ``"_predictions.tsv"``
        append : bool, default=False
            Should we append the current predictions to the file if it exists?
        class_labels : bool, default=True
            For classifier, should we convert class indices to their (str) labels
            for the returned array? Note that class labels are always written out
            to disk.
        individual_predictions : bool, default=False
            Return (and, optionally, write out) the predictions from each
            underlying learner.

        Returns
        -------
        Tuple[numpy.ndarray, Optional[Dict[str, numpy.ndarray]]]
            The first element is the array of predictions returned by the
            meta-estimator and the second is an optional dictionary with the
            name of each underlying learner as the key and the array of its
            predictions as the value. The second element is ``None`` if
            ``individual_predictions`` is set to ``False``.
        """
        example_ids = examples.ids

        # get the test set features in the right format
        if isinstance(examples.vectorizer, DictVectorizer):
            xtest = examples.vectorizer.inverse_transform(examples.features)

        # get all possible kinds of predictions from the meta-learner
        prediction_dict = get_predictions(self, xtest)

        # decide what predictions to write and what predictions to return
        # by default, these are just what is output by the model
        to_write = prediction_dict["raw"]
        to_return = prediction_dict["raw"]

        # for classifiers ...
        if self.learner_type == "classifier":
            # return and write class labels if they were explicitly asked for
            if class_labels:
                to_return = to_write = prediction_dict["labels"]
            else:
                # return and write probabilities
                if self.voting == "soft":
                    to_return = to_write = prediction_dict["probabilities"]
                # return class indices and write labels
                else:
                    to_return = prediction_dict["raw"]
                    to_write = prediction_dict["labels"]

        # for regressors, it's really simple
        else:
            to_write = to_return = prediction_dict["raw"]

        # write out the meta-estimator predictions if we are asked to
        if prediction_prefix is not None:
            write_predictions(
                example_ids,
                to_write,
                prediction_prefix,
                self.learner_type,
                self.learners[0].label_list,
                append=append,
            )

        # get and write each underlying learner's predictions if asked for
        if individual_predictions:
            # create a dictionary to hold the individual predictions
            individual_predictions_dict = {}

            # iterate over each underlying learner along with names
            for name, learner in zip(self.model.named_estimators_, self.learners):
                # the learner's `predict()` method should handle everything
                learner_prediction_prefix = (
                    f"{prediction_prefix}_{name}" if prediction_prefix is not None else None
                )
                learner_predictions = learner.predict(
                    examples,
                    prediction_prefix=learner_prediction_prefix,
                    append=append,
                    class_labels=class_labels,
                )

                # save this estimator's predictions in the dictionary
                individual_predictions_dict[name] = learner_predictions
        else:
            individual_predictions_dict = None

        # return the tuple of the meta-estimator predictions array
        # and the dictionary containing the individual predictions
        return (to_return, individual_predictions_dict)

    def evaluate(
        self,
        examples: FeatureSet,
        prediction_prefix: Optional[str] = None,
        append: bool = False,
        grid_objective: Optional[str] = None,
        individual_predictions: bool = False,
        output_metrics: List[str] = [],
    ) -> EvaluateTaskResults:
        """
        Evaluate the meta-estimator on a given ``FeatureSet``.

        Parameters
        ----------
        examples : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance to evaluate the performance of the model on.
        prediction_prefix : Optional[str], default=None
            If saving the predictions, this is the
            prefix that will be used for the filename.
            It will be followed by ``"_predictions.tsv"``
        append : bool, default=False
            Should we append the current predictions to the file if
            it exists?
        grid_objective : Optional[str], default=None
            The objective function used when doing the grid search.
        individual_predictions : bool, default=False
            Optionally, write out the predictions from each underlying learner.
        output_metrics : List[str], default=[]
            List of additional metric names to compute in
            addition to grid objective.

        Returns
        -------
        :class:`skll.types.EvaluateTaskResults`
            The confusion matrix, the overall accuracy, the per-label
            PRFs, the model parameters, the grid search objective
            function score, and the additional evaluation metrics, if any.
        """
        # make the prediction on the test data; note that these
        # are either class indices or class probabilities
        yhat, _ = self.predict(
            examples,
            class_labels=False,
            prediction_prefix=prediction_prefix,
            append=append,
            individual_predictions=individual_predictions,
        )

        # for classifiers, convert class labels indices for consistency
        # but account for any unseen labels in the test set that may not
        # have occurred in the training data for the underlying learners
        # at all; then get acceptable metrics based on the type of labels we have
        if examples.labels is not None:
            if self.learner_type == "classifier":
                sorted_unique_labels = np.unique(examples.labels)
                test_label_list = sorted_unique_labels.tolist()
                train_and_test_label_dict = add_unseen_labels(self.label_dict, test_label_list)
                ytest = np.array([train_and_test_label_dict[label] for label in examples.labels])
                acceptable_metrics = get_acceptable_classification_metrics(sorted_unique_labels)
            # for regressors we do not need to do anything special to the labels
            else:
                train_and_test_label_dict = None
                ytest = examples.labels
                acceptable_metrics = get_acceptable_regression_metrics()

        # check that all of the output metrics are acceptable
        unacceptable_metrics = set(output_metrics).difference(acceptable_metrics)
        if unacceptable_metrics and examples.labels is not None:
            label_type = examples.labels.dtype.type
            raise ValueError(
                f"The following metrics are not valid "
                f"for this learner({self.model_type.__name__}) "
                f"with these labels of type {label_type.__name__}: "
                f"{list(unacceptable_metrics)}"
            )

        # get the values of the evaluation metrics
        (
            conf_matrix,
            accuracy,
            result_dict,
            objective_score,
            metric_scores,
        ) = compute_evaluation_metrics(
            output_metrics,
            ytest,
            yhat,
            self.learner_type,
            label_dict=train_and_test_label_dict,
            grid_objective=grid_objective,
            probability=self.voting == "soft",
            logger=self.logger,
        )

        # add in the model parameters, excluding the ones
        # for the underlying estimators, and return
        model_params = self.model.get_params(deep=False)
        res = (conf_matrix, accuracy, result_dict, model_params, objective_score, metric_scores)
        return res

    def cross_validate(
        self,
        examples: FeatureSet,
        stratified: bool = True,
        cv_folds: Union[int, FoldMapping] = 10,
        cv_seed: int = 123456789,
        grid_search: bool = True,
        grid_search_folds: Union[int, FoldMapping] = 5,
        grid_jobs: Optional[int] = None,
        grid_objective: Optional[str] = None,
        output_metrics: List[str] = [],
        prediction_prefix: Optional[str] = None,
        param_grid_list: Optional[List[Dict[str, Any]]] = None,
        shuffle: bool = False,
        save_cv_folds: bool = True,
        save_cv_models: bool = False,
        individual_predictions: bool = False,
        use_custom_folds_for_grid_search: bool = True,
    ) -> VotingCrossValidateTaskResults:
        """
        Cross-validate the meta-estimator on the given examples.

        We follow essentially the same methodology as in
        ``Learner.cross_validate()`` - split the examples into
        training and testing folds, and then call ``self.train()``
        on the training folds and then ``self.evaluate()`` on the
        test fold. Note that this means that underlying estimators
        with different hyperparameters may be used for each fold, as is
        the case with ``Learner.cross_validate()``.

        Parameters
        ----------
        examples : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance to cross-validate learner performance on.
        stratified : bool, default=True
            Should we stratify the folds to ensure an even
            distribution of labels for each fold?
        cv_folds : Union[int, :class:`skll.types.FoldMapping`], default=10
            The number of folds to use for cross-validation, or
            a mapping from example IDs to folds.
        cv_seed: int, default=123456789
            The value for seeding the random number generator
            used to create the random folds. Note that this
            seed is *only* used if either ``grid_search`` or
            ``shuffle`` are set to ``True``.
        grid_search : bool, default=True
            Should we do grid search when training each fold?
            Note: This will make this take *much* longer.
        grid_search_folds : Union[int, :class:`skll.types.FoldMapping`], default=5
            The number of folds to use when doing the grid search, or a mapping
            from example IDs to folds.
        grid_jobs : Optional[int], default=None
            The number of jobs to run in parallel when doing the grid search.
            If ``None`` or 0, the number of grid search folds will be used.
        grid_objective : Optional[str], default=None
            The name of the objective function to use when doing the grid search.
            Must be specified if ``grid_search`` is ``True``.
        output_metrics : Optional[List[str]], default=[]
            List of additional metric names to compute in addition to the metric
            used for grid search.
        prediction_prefix : Optional[str], default=None
            If saving the predictions, this is the prefix that will be used for
            the filename. It will be followed by ``"_predictions.tsv"``
        param_grid_list : Optional[List[Dict[str, Any]]], default=None
            The list of parameters grid to search through for grid
            search, one for each underlying learner. The order of
            the dictionaries should correspond to the order If ``None``,
            the default parameter grids will be used for the underlying
            estimators.
        shuffle : bool, default=False
            Shuffle examples before splitting into folds for CV.
        save_cv_folds : bool, default=True
             Whether to save the cv fold ids or not?
        save_cv_models : bool, default=False
            Whether to save the cv models or not?
        individual_predictions : bool, default=False
            Write out the cross-validated predictions from each underlying
            learner as well.
        use_custom_folds_for_grid_search : bool, default=True
            If ``cv_folds`` is a custom dictionary, but ``grid_search_folds``
            is not, perhaps due to user oversight, should the same custom
            dictionary automatically be used for the inner grid-search
            cross-validation?

        Returns
        -------
        :class:`skll.types.CrossValidateTaskResults`
           A 3-tuple containing the following:

            List[:class:`skll.types.EvaluateTaskResults`]: the confusion matrix, overall accuracy,
            per-label PRFs, model parameters, objective function score, and
            evaluation metrics (if any) for each fold.

            Optional[:class:`skll.types.FoldMapping`]: dictionary containing the test-fold number
            for each id if ``save_cv_folds`` is ``True``, otherwise ``None``.

            Optional[List[:class:`skll.learner.voting.VotingLearner`]]: list of voting
            learners, one for each fold if ``save_cv_models`` is ``True``,
            otherwise ``None``.

        Raises
        ------
        ValueError
            If classification labels are not properly encoded as strings.
        ValueError
            If ``grid_search`` is ``True`` but ``grid_objective`` is ``None``.
        """
        # Seed the random number generator so that randomized algorithms are
        # replicable.
        random_state = np.random.RandomState(cv_seed)

        # We need to check whether the labels in the featureset are labels
        # or continuous values. If it's the latter, we need to raise an
        # an exception since the stratified splitting in sklearn does not
        # work with continuous labels. Note that although using random folds
        # _will_ work, we want to raise an error in general since it's better
        # to encode the labels as strings anyway for classification problems.
        if self.learner_type == "classifier" and type_of_target(examples.labels) not in [
            "binary",
            "multiclass",
        ]:
            raise ValueError(
                "Floating point labels must be encoded as strings for cross-validation."
            )

        # check that we have an objective since grid search is on by default
        # Note that `train()` would raise this error anyway later but it's
        # better to raise this early on so rather than after a whole bunch of
        # stuff has happened
        if grid_search and not grid_objective:
            raise ValueError(
                "Grid search is on by default. You must "
                "either specify a grid objective or turn off "
                "grid search."
            )

        # Shuffle so that the folds are random for the inner grid search CV.
        # If grid search is True but shuffle isn't, shuffle anyway.
        # You can't shuffle a scipy sparse matrix in place, so unfortunately
        # we make a copy of everything (and then get rid of the old version)
        if grid_search or shuffle:
            if grid_search and not shuffle:
                self.logger.warning(
                    "Training data will be shuffled to randomize "
                    "grid search folds.  Shuffling may yield "
                    "different results compared to scikit-learn."
                )
            ids, labels, features = sk_shuffle(
                examples.ids, examples.labels, examples.features, random_state=random_state
            )
            examples = FeatureSet(
                examples.name, ids, labels=labels, features=features, vectorizer=examples.vectorizer
            )

        # Call some setup code which will properly initialize the underlying
        # learners before they are eventually trained
        self._setup_underlying_learners(examples)

        # Set up the cross-validation iterator.
        kfold, cv_groups = setup_cv_fold_iterator(
            cv_folds, examples, self.learner_type, stratified=stratified, logger=self.logger
        )

        # When using custom CV folds (a dictionary), if we are planning to do
        # grid search, set the grid search folds to be the same as the custom
        # cv folds unless a flag is set that explicitly tells us not to.
        # Note that this should only happen when we are using the API; otherwise
        # the configparser should take care of this even before this method is called
        if isinstance(cv_folds, dict):
            if grid_search and use_custom_folds_for_grid_search and grid_search_folds != cv_folds:
                self.logger.warning(
                    "The specified custom folds will be used for " "the inner grid search."
                )
                grid_search_folds = cv_folds

        # handle each fold separately & accumulate the predictions and results
        results = []
        append_predictions = False
        saved_models: List["VotingLearner"] = []
        saved_skll_fold_ids: FoldMapping = {}
        if examples.features is not None and examples.labels is not None:
            for fold_num, (train_indices, test_indices) in enumerate(
                kfold.split(examples.features, examples.labels, cv_groups)
            ):
                # Train model
                self._model = None  # prevent feature vectorizer from being reset.
                train_set = FeatureSet(
                    examples.name,
                    examples.ids[train_indices],
                    labels=examples.labels[train_indices],
                    features=examples.features[train_indices],
                    vectorizer=examples.vectorizer,
                )

                self.train(
                    train_set,
                    param_grid_list=param_grid_list,
                    grid_search_folds=grid_search_folds,
                    grid_search=grid_search,
                    grid_objective=grid_objective,
                    grid_jobs=grid_jobs,
                    shuffle=grid_search,
                )

                if save_cv_models:
                    saved_models.append(copy.deepcopy(self))

                # evaluate the voting meta-estimator on the test fold
                test_tuple = FeatureSet(
                    examples.name,
                    examples.ids[test_indices],
                    labels=examples.labels[test_indices],
                    features=examples.features[test_indices],
                    vectorizer=examples.vectorizer,
                )

                # save the results
                results.append(
                    self.evaluate(
                        test_tuple,
                        prediction_prefix=prediction_prefix,
                        append=append_predictions,
                        grid_objective=grid_objective,
                        output_metrics=output_metrics,
                        individual_predictions=individual_predictions,
                    )
                )
                append_predictions = True

                # save the fold number for each test ID if we were asked to
                if save_cv_folds:
                    for index in test_indices:
                        saved_skll_fold_ids[examples.ids[index]] = str(fold_num)

        # return list of results/outputs for all folds
        models = saved_models if save_cv_models else None
        skll_fold_ids = saved_skll_fold_ids if save_cv_folds else None
        return (results, skll_fold_ids, models)

    def learning_curve(
        self,
        examples: FeatureSet,
        metric: str,
        cv_folds: Union[int, FoldMapping] = 10,
        train_sizes: LearningCurveSizes = np.linspace(0.1, 1.0, 5),
        override_minimum: bool = False,
    ) -> Tuple[List[float], List[float], List[float], List[int]]:
        """
        Generate learning curves for the meta-estimator.

        Generate learning curves for the voting meta-estimator on the training
        examples via cross-validation. Adapted from the scikit-learn code for
        learning curve generation (cf.``sklearn.model_selection.learning_curve``).

        Parameters
        ----------
        examples : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance to generate the learning curve on.
        metric : str
            The name of the metric function to use
            when computing the train and test scores
            for the learning curve.
        cv_folds : Union[int, :class:`skll.types.FoldMapping`], default=10
            The number of folds to use for cross-validation, or
            a mapping from example IDs to folds.
        train_sizes : :class:`skll.types.LearningCurveSizes`, default= :func:`numpy.linspace` with start=0.1, stop=1.0, num=5
            Relative or absolute numbers of training examples
            that will be used to generate the learning curve.
            If the type is float, it is regarded as a fraction
            of the maximum size of the training set (that is
            determined by the selected validation method),
            i.e. it has to be within (0, 1]. Otherwise it
            is interpreted as absolute sizes of the training
            sets. Note that for classification the number of
            samples usually have to be big enough to contain
            at least one sample from each class.
        override_minimum : bool, default=False
            Learning curves can be unreliable for very small sizes
            esp. for > 2 labels. If this option is set to ``True``, the
            learning curve would be generated even if the number
            of example is less 500 along with a warning. If ``False``,
            the curve is not generated and an exception is raised instead.

        Returns
        -------
        train_scores : List[float]
            The scores for the training set.
        test_scores : List[float]
            The scores on the test set.
        fit_times : List[float]
            The average times taken to fit each model.
        num_examples : List[int]
            The numbers of training examples used to generate the curve.

        Raises
        ------
        ValueError
            If the number of examples is less than 500.
        """
        # check that the number of training examples is more than the minimum
        # needed for generating a reliable learning curve
        if len(examples) < 500:
            if not override_minimum:
                raise ValueError(
                    f"Number of training examples provided ({len(examples)}) "
                    "is less than the minimum needed (500) for the "
                    "learning curve to be reliable."
                )
            else:
                self.logger.warning(
                    "Learning curves can be unreliable for examples fewer than "
                    f"500. You provided {len(examples)}."
                )

        # raise a warning if we are using a probabilistic classifier
        # since that means we cannot use the predictions directly
        if self.voting == "soft":
            self.logger.warning(
                "For soft-voting classifiers, the most likely "
                "class will be computed via an argmax before "
                "computing the curve."
            )

        # Call some setup code which will properly initialize the underlying
        # learners before they are eventually trained
        self._setup_underlying_learners(examples)

        # set up the CV split iterator over the train/test featuresets
        # which also returns the maximum number of training examples
        (featureset_iter, n_max_training_samples) = setup_cv_split_iterator(cv_folds, examples)

        # Get the `_translate_train_sizes()` function from scikit-learn
        # since we need it to get the right list of sizes after cross-validation
        _module = import_module("sklearn.model_selection._validation")
        _translate_train_sizes = getattr(_module, "_translate_train_sizes")
        train_sizes_abs = _translate_train_sizes(train_sizes, n_max_training_samples)
        n_unique_ticks = train_sizes_abs.shape[0]

        # limit the number of parallel jobs for this to be no higher than
        # MAX_CONCURRENT_PROCESSES or the number of cores, whichever is lower
        n_jobs = min(cpu_count(), MAX_CONCURRENT_PROCESSES)

        # Run jobs in parallel that train the model on each subset
        # of the training data and compute train and test scores
        parallel = joblib.Parallel(n_jobs=n_jobs, pre_dispatch=n_jobs)
        out = parallel(
            joblib.delayed(train_and_score)(self, train_fs[:n_train_samples], test_fs, metric)
            for train_fs, test_fs in featureset_iter
            for n_train_samples in train_sizes_abs
        )

        # Reshape the outputs
        out = np.array(out)
        n_cv_folds = out.shape[0] // n_unique_ticks
        out = out.reshape(n_cv_folds, n_unique_ticks, 3)
        out = np.asarray(out).transpose((2, 1, 0))

        return list(out[0]), list(out[1]), list(out[2]), list(train_sizes_abs)



#!/usr/bin/env python
# License: BSD 3 clause
"""
Script that joins a bunch of feature files together to create one file.

:author: Dan Blanchard (dblanchard@ets.org)
:organization: ETS
"""

import argparse
import logging
import sys
from pathlib import Path
from typing import List, Optional

from skll.data.readers import EXT_TO_READER
from skll.data.writers import EXT_TO_WRITER, ARFFWriter, CSVWriter, TSVWriter
from skll.version import __version__


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and gets things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Joins multiple input feature files together into one " " file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "infile",
        help="input feature file (ends in .arff, .csv, " " .jsonlines, .ndj, or .tsv)",
        nargs="+",
    )
    parser.add_argument("outfile", help="output feature file")
    parser.add_argument(
        "-i",
        "--id_col",
        help="Name of the column which contains the instance " " IDs in ARFF, CSV, or TSV files.",
        default="id",
    )
    parser.add_argument(
        "-l",
        "--label_col",
        help="Name of the column which contains the class "
        "labels in ARFF, CSV, or TSV files. For ARFF "
        "files, this must be the final column to count "
        "as the label.",
        default="y",
    )
    parser.add_argument(
        "-q", "--quiet", help='Suppress printing of "Loading..." messages.', action="store_true"
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s")
    logger = logging.getLogger(__name__)

    # all extensions except .libsvm can be processed
    valid_extensions = {ext for ext in EXT_TO_READER if ext != ".libsvm"}

    # make sure the input file extensions are those we can process
    input_extension_set = {Path(inf).suffix.lower() for inf in args.infile}
    output_extension = Path(args.outfile).suffix.lower()

    # make sure all the files are in the same format except libsvm files
    if len(input_extension_set) > 1:
        logger.error("All input files must be in the same format.")
        sys.exit(1)

    input_extension = list(input_extension_set)[0]

    if input_extension not in valid_extensions:
        logger.error(
            "Input files must be in either .arff, .csv, .jsonlines, "
            f".ndj, or .tsv format. You specified: {input_extension}"
        )
        sys.exit(1)

    if output_extension != input_extension:
        logger.error(
            "Output file must be in the same format as the input "
            f"file.  You specified: {output_extension}"
        )
        sys.exit(1)

    # Read and merge input files
    merged_set = None
    for infile in args.infile:
        reader = EXT_TO_READER[input_extension](
            infile, quiet=args.quiet, label_col=args.label_col, id_col=args.id_col
        )
        fs = reader.read()
        if merged_set is None:
            merged_set = fs
        else:
            merged_set += fs

    # write out the file in the requested output format
    writer_type = EXT_TO_WRITER[input_extension]
    writer_args = {"quiet": args.quiet}
    if writer_type is CSVWriter or writer_type is TSVWriter:
        writer_args["label_col"] = args.label_col
        writer_args["id_col"] = args.id_col
    elif writer_type is ARFFWriter:
        writer_args["label_col"] = args.label_col
        writer_args["id_col"] = args.id_col
        writer_args["regression"] = reader.regression
        writer_args["relation"] = reader.relation
    writer = writer_type(args.outfile, merged_set, **writer_args)
    writer.write()


if __name__ == "__main__":
    main()



# License: BSD 3 clause
"""
Handles loading data from various types of data files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:author: Jeremy Biggs (jbiggs@ets.org)
:organization: ETS
"""

import json
import logging
import sys
from csv import DictWriter
from decimal import Decimal
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
from scipy.sparse import issparse
from sklearn.feature_extraction import DictVectorizer, FeatureHasher

from skll.data import FeatureSet
from skll.types import FeatGenerator, FeatureDict, IdType, LabelType, PathOrStr


class Writer(object):
    """
    Write out FeatureSets to files on disk.

    This is the base class used to create featureset writers for different
    file types.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        A path to the feature file we would like to create. The suffix
        to this filename must be ``.arff``, ``.csv``, ``.jsonlines``,
        ``.libsvm``, ``.ndj``, or ``.tsv``. If ``subsets``
        is not ``None``, when calling the ``write()`` method, path is
        assumed to be a string containing the path to the directory to
        write the feature files with an additional file extension
        specifying the file type. For example ``/foo/.csv``.

    feature_set : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance to dump to the file.

    quiet : bool, default=True
        Do not print "Writing..." status message to stderr.

    subsets : Optional[Dict[str, List[str]]], default=None
        A mapping from subset names to lists of feature names
        that are included in those sets. If given, a feature
        file will be written for every subset (with the name
        containing the subset name as suffix to ``path``).
        Note, since string- valued features are automatically
        converted into boolean features with names of the form
        ``FEATURE_NAME=STRING_VALUE``, when doing the
        filtering, the portion before the ``=`` is all that's
        used for matching. Therefore, you do not need to
        enumerate all of these boolean feature names in your
        mapping.

    logger : Optional[logging.Logger], default=None
        A logger instance to use to log messages instead of creating
        a new one by default.
    """

    def __init__(
        self,
        path: PathOrStr,
        feature_set: FeatureSet,
        quiet: bool = True,
        subsets: Optional[Dict[str, List[str]]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize base Writer class."""
        super(Writer, self).__init__()

        self.quiet = quiet
        self.path = Path(path)
        self.feat_set = feature_set
        self.subsets = subsets
        self.logger = logger if logger else logging.getLogger(__name__)

        # Get prefix & extension for checking file types & writing subset files;
        # since we also need to handle paths like "foo/.csv" for subset-writing
        # we have to do a bit of introspection before figuring out the various
        # parts of the path
        parent, stem, suffix = self.path.parent, self.path.stem, self.path.suffix
        if stem.startswith(".") and not suffix:
            self.root = parent
            self.ext = stem.lower()
        else:
            self.root = parent / stem
            self.ext = suffix.lower()
        self._progress_msg = ""
        self._use_pandas = False

    @classmethod
    def for_path(cls, path: PathOrStr, feature_set: FeatureSet, **kwargs) -> "Writer":
        """
        Retrieve object of ``Writer`` sub-class appropriate for given path.

        Parameters
        ----------
        path : :class:`skll.types.PathOrStr`
            A path to the feature file we would like to create. The
            suffix to this filename must be ``.arff``, ``.csv``,
            ``.jsonlines``, ``.libsvm``, ``.ndj``, or
            ``.tsv``. If ``subsets`` is not ``None``, when calling the
            ``write()`` method, path is assumed to be a string
            containing the path to the directory to write the feature
            files with an additional file extension specifying the
            file type. For example ``/foo/.csv``.

        feature_set : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance to dump to the output file.

        kwargs : Optional[Dict[str, Any]]
            The keyword arguments for ``for_path`` are the same as
            the initializer for the desired ``Writer`` subclass.

        Returns
        -------
        writer : :class:`skll.data.Writer`
            New instance of the Writer sub-class that is
            appropriate for the given path.
        """
        # Get lowercase extension for file extension checking
        # NOTE: the reason we are doing this complicated gymnastics
        # instead of just using `path.suffix` is because sometimes
        # `path` may look like `foo/.jsonlines` when we are writing
        # subsets so we need to handle that edge case
        path = Path(path)
        stem, suffix = path.stem, path.suffix
        if stem.startswith(".") and not suffix:
            ext = stem.lower()
        else:
            ext = suffix.lower()
        return EXT_TO_WRITER[ext](path, feature_set, **kwargs)

    def write(self) -> None:
        """Write out this Writer's ``FeatureSet`` to a file in its format."""
        if isinstance(self.feat_set.vectorizer, FeatureHasher):
            raise ValueError(
                "Writer cannot write sets that use a " "FeatureHasher for vectorization."
            )

        # Write one feature file if we weren't given a dict of subsets
        if self.subsets is None:
            self._write_subset(self.path)

        # Otherwise write one feature file per subset
        else:
            for subset_name, filter_features in self.subsets.items():
                self.logger.debug(f"Subset ({subset_name}) features: " f"{filter_features}")
                sub_path = self.root / f"{subset_name}{self.ext}"
                self._write_subset(sub_path, set(filter_features))

    def _write_subset(
        self, sub_path: PathOrStr, filter_features: Optional[Set[str]] = None
    ) -> None:
        """
        Write out given ``FeatureSet`` instance to a file in this class's format.

        Parameters
        ----------
        sub_path : :class:`skll.types.PathOrStr`
            The path to the file we want to create for this subset
            of our data.

        filter_features : Optional[Set[str]], default=None
            Set of features to include in current feature file.
        """
        self.logger.debug(f"sub_path: {sub_path}")
        self.logger.debug(f"feature_set: {self.feat_set.name}")
        self.logger.debug(f"filter_features: {filter_features}")

        if not self.quiet:
            self._progress_msg = f"Writing {sub_path}..."
            print(self._progress_msg, end="\r", file=sys.stderr)
            sys.stderr.flush()

        if not self._use_pandas:
            # Apply filtering
            filtered_set: Union[FeatGenerator, FeatureSet] = (
                self.feat_set.filtered_iter(features=filter_features)
                if filter_features is not None
                else self.feat_set
            )

            # Open file for writing and write each line
            with open(sub_path, "w", encoding="utf-8") as output_file:
                # Write out the header if this format requires it
                self._write_header(filtered_set, output_file, filter_features)
                # Write individual lines
                for ex_num, (id_, label_, feat_dict) in enumerate(filtered_set):
                    self._write_line(id_, label_, feat_dict, output_file)
                    if not self.quiet and ex_num % 100 == 0:
                        print(f"{self._progress_msg}{ex_num:>15}", end="\r", file=sys.stderr)
                        sys.stderr.flush()
        else:
            self._write_data(self.feat_set, sub_path, filter_features)

        if not self.quiet:
            print(f"{self._progress_msg}{'done':<15}", file=sys.stderr)
            sys.stderr.flush()

    def _write_header(self, feature_set, output_file, filter_features):
        """
        Write header to file.

        Called before lines are written to file, so that headers can be written
        for files that need them.

        Parameters
        ----------
        feature_set : Ignored
            Not used.

        output_file : Ignored
            Not used.

        filter_features : Ignored
           Not used.
        """
        pass

    def _write_line(self, id_, label_, feat_dict, output_file):
        """
        Write the current line in the file in this Writer's format.

        Parameters
        ----------
        id_ : Ignored
            Not used.

        label_ : Ignored
            Not used.

        feat_dict : Ignored
            Not used.

        output_file : Ignored
             Not used.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def _write_data(self, feature_set, output_file, filter_features):
        """
        Write full data set in Writer's format using `pandas`, rather than row-by-row.

        Parameters
        ----------
        feature_set : Ignored
            Not used.

        output_file : Ignored
            Not used.

        filter_features : Ignored
            Not used.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def _get_column_names_and_indexes(
        self, feature_set: FeatureSet, filter_features: Optional[Set[str]] = None
    ) -> Tuple[List[str], List[int]]:
        """
        Get names of columns and associated indices for (possibly filtered) features.

        Parameters
        ----------
        feature_set : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance being written to a file.

        filter_features : Optional[Set[str]], default=None
            If only writing a subset of the features in the
            FeatureSet to ``output_file``, these are the
            features to include in this file.

        Returns
        -------
        column_names : List[str]
            A list of the (possibly filtered) column names.

        column_indexes : List[int]
            A list of the (possibly filtered) column indexes.
        """
        # if we're not doing filtering,
        # then just take all the feature names
        self.logger.debug(feature_set)
        if isinstance(feature_set.vectorizer, DictVectorizer):
            if filter_features is None:
                filter_features = feature_set.vectorizer.feature_names_

            # create a list of tuples with (column names, column indexes)
            # so that we can correctly extract the appropriate columns
            columns = sorted(
                [
                    (col_name, col_idx)
                    for col_name, col_idx in feature_set.vectorizer.vocabulary_.items()
                    if (col_name in filter_features or col_name.split("=", 1)[0] in filter_features)
                ],
                key=lambda x: x[1],
            )

            # then, split the names and indexes into separate lists
            column_names, column_indexes = zip(*columns)
            return list(column_names), list(column_indexes)
        else:
            return [], []


class CSVWriter(Writer):
    """
    Writer for writing out ``FeatureSet`` instances as CSV files.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        A path to the feature file we would like to create.
        If ``subsets`` is not ``None``, this is assumed to be a string
        containing the path to the directory to write the feature
        files with an additional file extension specifying the file
        type. For example ``/foo/.csv``.

    feature_set : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance to dump to the output file.

    quiet : bool, default=True
        Do not print "Writing..." status message to stderr.

    subsets : Optional[Dict[str, List[str]]], default=None
        A mapping from subset names to lists of feature names
        that are included in those sets. If given, a feature
        file will be written for every subset (with the name
        containing the subset name as suffix to ``path``).
        Note, since string- valued features are automatically
        converted into boolean features with names of the form
        ``FEATURE_NAME=STRING_VALUE``, when doing the
        filtering, the portion before the ``=`` is all that's
        used for matching. Therefore, you do not need to
        enumerate all of these boolean feature names in your
        mapping.

    logger : Optional[logging.Logger], default=None
        A logger instance to use to log messages instead of creating
        a new one by default.

    label_col : str, default="y"
        The column name containing the label.

    id_col : str, default="id"
        The column name containing the ID.

    pandas_kwargs : Optional[Dict[str], Any], default=None
        Arguments that will be passed directly to the `pandas` I/O reader.
    """

    def __init__(
        self,
        path: PathOrStr,
        feature_set: FeatureSet,
        quiet: bool = True,
        subsets: Optional[Dict[str, List[str]]] = None,
        logger: Optional[logging.Logger] = None,
        label_col: str = "y",
        id_col: str = "id",
        pandas_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the CSVWriter class."""
        self.label_col = label_col
        self.id_col = id_col
        self._pandas_kwargs = {} if pandas_kwargs is None else pandas_kwargs
        self._sep = self._pandas_kwargs.pop("sep", ",")
        self._index = self._pandas_kwargs.pop("index", False)
        super(CSVWriter, self).__init__(
            path, feature_set, quiet=quiet, subsets=subsets, logger=logger
        )
        self._use_pandas = True

    def _build_dataframe_with_features(
        self, feature_set: FeatureSet, filter_features: Optional[Set[str]] = None
    ) -> pd.DataFrame:
        """
        Create and filter data frame from features in given feature set.

        Parameters
        ----------
        feature_set : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance being written to a file.

        filter_features : Optional[Set[str]], default=None
            If only writing a subset of the features in the
            FeatureSet to ``output_file``, these are the
            features to include in this file.

        Returns
        -------
        df_features : pandas.DataFrame
            The data frame constructed from the feature set. The frame may be
            empty are not features in the feature set.

        Raises
        ------
        ValueError
            If ID column is already used as feature.
            If label column is already used as feature.
        """
        # if there is no filtering, then just keep all the names
        (column_names, column_idxs) = self._get_column_names_and_indexes(
            feature_set, filter_features
        )

        # create the data frame from the feature set;
        # then, select only the columns that we want,
        # and give the columns their correct names
        if feature_set.features is not None:
            if issparse(feature_set.features):
                df_features = pd.DataFrame(feature_set.features.toarray())
            else:
                df_features = pd.DataFrame(feature_set.features)
            df_features = df_features.iloc[:, column_idxs].copy()
            df_features.columns = column_names
            return df_features
        else:
            return pd.DataFrame()

    def _build_dataframe(
        self,
        feature_set: FeatureSet,
        filter_features: Optional[Set[str]] = None,
        df_features: Optional[pd.DataFrame] = None,
    ) -> pd.DataFrame:
        """
        Create and filter data frame with features in given feature set.

        Add the IDs and labels, if applicable. If the data frame
        with features already exists, pass `df_features`. Then the IDs and labels will
        simply be added to the existing data frame containing the features.

        Parameters
        ----------
        feature_set : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance being written to a file.

        filter_features : Optional[Set[str]], default=None
            If only writing a subset of the features in the
            FeatureSet to ``output_file``, these are the
            features to include in this file.

        df_features : Optional[pandas.DataFrame], default=None
            If the data frame with features already exists,
            then we use it and add IDs and labels; otherwise,
            the feature data frame will be created from the feature set.

        Returns
        -------
        df_features : pandas.DataFrame
            The data frame constructed from the feature set.

        Raises
        ------
        ValueError
            If ID column is already used as feature.
            If label column is already used as feature.
        """
        # create the data frame with just the features
        # from the feature set, at this point
        if df_features is None:
            df_features = self._build_dataframe_with_features(feature_set, filter_features)

        # if the id column is already in the data frame,
        # then raise an error; otherwise, just add the ids
        if self.id_col in df_features:
            raise ValueError(f"ID column name {self.id_col} already used as feature name.")
        df_features[self.id_col] = feature_set.ids

        # if the the labels should exist but the column is already
        # in the data frame, then raise an error; otherwise, just add the labels
        if feature_set.has_labels:
            if self.label_col in df_features:
                raise ValueError(
                    f'Class column name "{self.label_col}" ' "already used as feature name."
                )
            df_features[self.label_col] = feature_set.labels

        return df_features

    def _write_data(
        self,
        feature_set: FeatureSet,
        output_file: PathOrStr,
        filter_features: Optional[Set[str]] = None,
    ) -> None:
        """
        Write the data in CSV format.

        Parameters
        ----------
        feature_set : :class:`skll.data.featureset.FeatureSet`
            The ``FeatureSet`` instance being written to a file.

        output_file : :class:`skll.types.PathOrStr`
            The path of the file being written to

        filter_features : Optional[Set[str]], default=None
            If only writing a subset of the features in the
            FeatureSet to ``output_file``, these are the
            features to include in this file.
        """
        df = self._build_dataframe(feature_set, filter_features=filter_features)
        df.to_csv(output_file, sep=self._sep, index=self._index, **self._pandas_kwargs)


class TSVWriter(CSVWriter):
    """
    Writer for writing out FeatureSets as TSV files.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        A path to the feature file we would like to create.
        If ``subsets`` is not ``None``, this is assumed to be a string
        containing the path to the directory to write the feature
        files with an additional file extension specifying the file
        type. For example ``/foo/.tsv``.

    feature_set : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance to dump to the output file.

    quiet : bool, default=True
        Do not print "Writing..." status message to stderr.

    subsets : Optional[Dict[str, List[str]]], default=None
        A mapping from subset names to lists of feature names
        that are included in those sets. If given, a feature
        file will be written for every subset (with the name
        containing the subset name as suffix to ``path``).
        Note, since string- valued features are automatically
        converted into boolean features with names of the form
        ``FEATURE_NAME=STRING_VALUE``, when doing the
        filtering, the portion before the ``=`` is all that's
        used for matching. Therefore, you do not need to
        enumerate all of these boolean feature names in your
        mapping.

    logger : Optional[logging.Logger], default=None
        A logger instance to use to log messages instead of creating
        a new one by default.

    label_col: str, default="y"
        The column name containing the label.

    id_col: str, default="id"
        The column name containing the ID.

    pandas_kwargs : Optional[Dict[str, Any]], default=None
        Arguments that will be passed directly to the `pandas` I/O reader.
    """

    def __init__(
        self,
        path: PathOrStr,
        feature_set: FeatureSet,
        quiet: bool = True,
        subsets: Optional[Dict[str, List[str]]] = None,
        logger: Optional[logging.Logger] = None,
        label_col: str = "y",
        id_col: str = "id",
        pandas_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the TSVWriter class."""
        super(TSVWriter, self).__init__(
            path,
            feature_set,
            quiet=quiet,
            subsets=subsets,
            logger=logger,
            label_col=label_col,
            id_col=id_col,
            pandas_kwargs=pandas_kwargs,
        )
        self._sep = str("\t")


class ARFFWriter(Writer):
    """
    Writer for writing out FeatureSets as ARFF files.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        A path to the feature file we would like to create.
        If ``subsets`` is not ``None``, this is assumed to be a string
        containing the path to the directory to write the feature
        files with an additional file extension specifying the file
        type. For example ``/foo/.arff``.

    feature_set : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance to dump to the output file.

    quiet : bool, default=True
        Do not print "Writing..." status message to stderr.

    subsets : Optional[Dict[str, List[str]]], default=None
        A mapping from subset names to lists of feature names
        that are included in those sets. If given, a feature
        file will be written for every subset (with the name
        containing the subset name as suffix to ``path``).
        Note, since string- valued features are automatically
        converted into boolean features with names of the form
        ``FEATURE_NAME=STRING_VALUE``, when doing the
        filtering, the portion before the ``=`` is all that's
        used for matching. Therefore, you do not need to
        enumerate all of these boolean feature names in your
        mapping.

    logger : Optional[logging.Logger], default=None
        A logger instance to use to log messages instead of creating
        a new one by default.

    relation : str, default='skll_relation'
        The name of the relation in the ARFF file.

    regression : bool, default=False
        Is this an ARFF file to be used for regression?

    kwargs : Optional[Dict[str, Any]]
        The arguments to the ``Writer`` object being instantiated.
    """

    def __init__(
        self,
        path: PathOrStr,
        feature_set: FeatureSet,
        quiet: bool = True,
        subsets: Optional[Dict[str, List[str]]] = None,
        logger: Optional[logging.Logger] = None,
        relation="skll_relation",
        regression=False,
        dialect="excel-tab",
        label_col="y",
        id_col="id",
    ):
        """Initialize the ARFFWRiter class."""
        self.relation = relation
        self.regression = regression
        self.dialect = dialect
        self.label_col = label_col
        self.id_col = id_col
        super(ARFFWriter, self).__init__(
            path, feature_set, quiet=quiet, subsets=subsets, logger=logger
        )
        self._dict_writer: Optional[DictWriter[str]] = None

    def _write_header(
        self, feature_set: FeatureSet, output_file: IO[str], filter_features: Set[str]
    ) -> None:
        """
        Write headers to ARFF file.

        Called before lines are written to file, so that headers can be written
        for files that need them.

        Parameters
        ----------
        feature_set : Ignored
            Not used.

        output_file : IO[str]
            The file being written to.

        filter_features : Set[str]
            If only writing a subset of the features in the
            FeatureSet to ``output_file``, these are the
            features to include in this file.
        """
        fieldnames, _ = self._get_column_names_and_indexes(self.feat_set, filter_features)
        fieldnames.append(self.id_col)

        # Add relation to header
        print(f"@relation '{self.relation}'\n", file=output_file)

        # Loop through fields writing the header info for the ARFF file
        for field in fieldnames:
            field = field.replace("\\", "\\\\").replace("'", "\\'")
            print(f"@attribute '{field}' numeric", file=output_file)

        # Print class label header if necessary
        if self.regression:
            print(f"@attribute {self.label_col} numeric", file=output_file)
        else:
            if self.feat_set.has_labels:
                sorted_features = sorted(set(self.feat_set.labels))  # type: ignore
                labels_str = ",".join([str(feat) for feat in sorted_features])
                labels_str = "{" + labels_str + "}"
                print(f"@attribute {self.label_col} {labels_str}", file=output_file)
        if self.label_col:
            fieldnames.append(self.label_col)

        # Create CSV writer to handle missing values for lines in data section
        # and to ignore the instance values for non-numeric attributes
        self._dict_writer = DictWriter(
            output_file, fieldnames, restval=0, extrasaction="ignore", dialect="arff"
        )

        # Finish header and start data section
        print("\n@data", file=output_file)

    def _write_line(
        self, id_: IdType, label_: LabelType, feat_dict: FeatureDict, output_file: IO[str]
    ) -> None:
        """
        Write the current line in the file in this Writer's format.

        Parameters
        ----------
        id_ : :class:`skll.types.IdType`
            The ID for the current instance.

        label_ : :class:`skll.types.LabelType`
            The label for the current instance.

        feat_dict : :class:`skll.types.FeatureDict`
            The feature dictionary for the current instance.

        output_file : Ignored.
            Not used.

        Raises
        ------
        ValueError
            If class column name is already use as a feature

        ValueError
            If ID column name is already used as a feature.
        """
        # Add class column to feat_dict (unless this is unlabeled data)
        if self.label_col not in feat_dict:
            if self.feat_set.has_labels:
                feat_dict[self.label_col] = label_
        else:
            raise ValueError(
                f'Class column name "{self.label_col}" already ' "used as feature name."
            )
        # Add id column to feat_dict if id is provided
        if self.id_col not in feat_dict:
            feat_dict[self.id_col] = id_
        else:
            raise ValueError(f'ID column name "{self.id_col}" already used as' " feature name.")

        # Write out line
        if self._dict_writer:
            self._dict_writer.writerow(feat_dict)


class NDJWriter(Writer):
    """
    Writer for writing out FeatureSets as .jsonlines/.ndj files.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        A path to the feature file we would like to create.
        If ``subsets`` is not ``None``, this is assumed to be a string
        containing the path to the directory to write the feature
        files with an additional file extension specifying the file
        type. For example ``/foo/.ndj``.

    feature_set : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance to dump to the output file.

    quiet : bool, default=True
        Do not print "Writing..." status message to stderr.

    subsets : Optional[Dict[str, List[str]]], default=None
        A mapping from subset names to lists of feature names
        that are included in those sets. If given, a feature
        file will be written for every subset (with the name
        containing the subset name as suffix to ``path``).
        Note, since string- valued features are automatically
        converted into boolean features with names of the form
        ``FEATURE_NAME=STRING_VALUE``, when doing the
        filtering, the portion before the ``=`` is all that's
        used for matching. Therefore, you do not need to
        enumerate all of these boolean feature names in your
        mapping.

    logger : Optional[logging.Logger], default=None
        A logger instance to use to log messages instead of creating
        a new one by default.
    """

    def __init__(
        self,
        path: PathOrStr,
        feature_set: FeatureSet,
        quiet: bool = True,
        subsets: Optional[Dict[str, List[str]]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize the NDJWriter class."""
        super(NDJWriter, self).__init__(
            path, feature_set, quiet=quiet, subsets=subsets, logger=logger
        )

    def _write_line(
        self,
        id_: IdType,
        label_: Union[LabelType, np.int64, np.float64],
        feat_dict: FeatureDict,
        output_file: IO[str],
    ) -> None:
        """
        Write the current line in the file in NDJ format.

        Parameters
        ----------
        id_ : :class:`skll.types.IdType`
            The ID for the current instance.

        label_ : :class:`skll.types.LabelType`
            The label for the current instance.

        feat_dict : :class:`skll.types.FeatureDict`
            The feature dictionary for the current instance.

        output_file : IO[str]
            The file being written to.
        """
        example_dict: FeatureDict = {}
        # Don't try to add class column if this is label-less data
        # Try to convert the label to a scalar assuming it'a numpy
        # non-scalar type (e.g., int64) but if that doesn't work
        # then use it as is
        if self.feat_set.has_labels:
            if hasattr(label_, "item"):
                example_dict["y"] = label_.item()
            else:
                example_dict["y"] = label_
        # Try to convert the ID to a scalar assuming it'a numpy
        # non-scalar type (e.g., int64) but if that doesn't work
        # then use it as is
        if hasattr(id_, "item"):
            example_dict["id"] = id_.item()
        else:
            example_dict["id"] = id_
        example_dict["x"] = feat_dict
        print(json.dumps(example_dict, sort_keys=True), file=output_file)


class LibSVMWriter(Writer):
    """
    Writer for writing out FeatureSets as LibSVM/SVMLight files.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        A path to the feature file we would like to create.
        If ``subsets`` is not ``None``, this is assumed to be a string
        containing the path to the directory to write the feature
        files with an additional file extension specifying the file
        type. For example ``/foo/.libsvm``.

    feature_set : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance to dump to the output file.

    quiet : bool, default=True
        Do not print "Writing..." status message to stderr.

    subsets : Optional[Dict[str, List[str]]], default=None
        A mapping from subset names to lists of feature names
        that are included in those sets. If given, a feature
        file will be written for every subset (with the name
        containing the subset name as suffix to ``path``).
        Note, since string- valued features are automatically
        converted into boolean features with names of the form
        ``FEATURE_NAME=STRING_VALUE``, when doing the
        filtering, the portion before the ``=`` is all that's
        used for matching. Therefore, you do not need to
        enumerate all of these boolean feature names in your
        mapping.

    logger : Optional[logging.Logger], default=None
        A logger instance to use to log messages instead of creating
        a new one by default.

    label_map : Optional[Dict[str, int]], default=None
        A mapping from label strings to integers.
    """

    LIBSVM_REPLACE_DICT = {
        ":": "\u2236",
        "#": "\uFF03",
        " ": "\u2002",
        "=": "\ua78a",
        "|": "\u2223",
    }

    def __init__(
        self,
        path: PathOrStr,
        feature_set: FeatureSet,
        quiet: bool = True,
        subsets: Optional[Dict[str, List[str]]] = None,
        logger: Optional[logging.Logger] = None,
        label_map: Optional[Dict[Any, Any]] = None,
    ):
        """Initialize the LibSVMWriter class."""
        self.label_map = label_map
        super(LibSVMWriter, self).__init__(
            path, feature_set, quiet=quiet, subsets=subsets, logger=logger
        )
        if self.label_map is None:
            fs_labels = feature_set.labels if feature_set.has_labels else np.array([])
            self.label_map = {
                label: num
                for num, label in enumerate(
                    sorted(
                        {
                            label
                            for label in fs_labels  # type: ignore
                            if not isinstance(label, (int, float))
                        }
                    )
                )
            }
            # Add fake item to vectorizer for None
            self.label_map[None] = "00000"

    @staticmethod
    def _sanitize(name: Union[IdType, LabelType]) -> Union[IdType, LabelType]:
        """
        Sanitize feature names for older feature formats.

        Replace special characters in names with close unicode
        equivalents to make things loadable in by LibSVM, LibLinear, or
        SVMLight.

        Parameters
        ----------
        name : Union[:class:`skll.types.IdType`, :class:`skll.types.LabelType`]
            Input name in which special characters are replaced with unicode
            equivalents.

        Returns
        -------
        Union[:class:`skll.types.IdType`, :class:`skll.types.LabelType`]
            The sanitized name with special characters replaced.
        """
        sanitized_name = name
        if isinstance(sanitized_name, str):
            for orig, replacement in LibSVMWriter.LIBSVM_REPLACE_DICT.items():
                sanitized_name = sanitized_name.replace(orig, replacement)
        return sanitized_name

    def _write_line(
        self, id_: IdType, label_: LabelType, feat_dict: FeatureDict, output_file: IO[str]
    ) -> None:
        """
        Write the current line in the file in this Writer's format.

        Parameters
        ----------
        id_ : :class:`skll.types.IdType`
            The ID for the current instance.

        label_ : :class:`skll.types.LabelType`
            The label for the current instance.

        feat_dict : :class:`skll.types.FeatureDict`
            The feature dictionary for the current instance.

        output_file : IO[str]
            The file being written to.
        """
        field_values = (
            sorted(
                [
                    (self.feat_set.vectorizer.vocabulary_[field] + 1, value)
                    for field, value in feat_dict.items()
                    if Decimal(value) != 0
                ]
            )
            if self.feat_set.vectorizer
            else []
        )

        # Print label
        if self.label_map:
            if label_ in self.label_map:
                print(self.label_map[label_], end=" ", file=output_file)
            else:
                print(label_, end=" ", file=output_file)

        # Print features
        print(
            " ".join((f"{field}:{value}" for field, value in field_values)),
            end=" ",
            file=output_file,
        )

        # Print comment with id and mappings
        print("#", end=" ", file=output_file)
        print(self._sanitize(id_), end="", file=output_file)
        print(" |", end=" ", file=output_file)

        if self.label_map:
            if label_ in self.label_map:
                print(
                    f"{self._sanitize(self.label_map[label_])}={self._sanitize(label_)}",
                    end=" | ",
                    file=output_file,
                )
            else:
                print(" |", end=" ", file=output_file)

        line = (
            " ".join(
                f"{self.feat_set.vectorizer.vocabulary_[field] + 1}={self._sanitize(field)}"
                for field, value in feat_dict.items()
                if Decimal(value) != 0
            )
            if self.feat_set.vectorizer
            else ""
        )
        print(line, file=output_file)


# Constants
EXT_TO_WRITER = {
    ".arff": ARFFWriter,
    ".csv": CSVWriter,
    ".jsonlines": NDJWriter,
    ".libsvm": LibSVMWriter,
    ".ndj": NDJWriter,
    ".tsv": TSVWriter,
}



# License: BSD 3 clause
"""
Define version number.

This module exists solely for version information so we only have to change it
in one place. Based on the suggestion `here. <http://bit.ly/16LbuJF>`_

:author: Dan Blanchard (dblanchard@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

__version__ = "5.0.0"
VERSION = tuple(int(x) for x in __version__.split("."))
